
//
// AXI decoder error
//
// Returns an error condition for all traffic
// Intended to cover unmapped address ranges
// (See also the axi splitter)
// It sinks all write data and returns zero read data with an error code.
// There is an additional 'fault' output which can be used as 
// an error condition outside the AXI bus standard
//
// On purpose the first read data is returned one cycle 
// after the address as not all master have been designed 
// to cope with immediate return data
// 
// 
// Design by G.J. van Loo, FenLogic Ltd, October-2017.
//
// This program is free software. It comes without any guarantees or
// warranty to the extent permitted by applicable law. Although the
// author has attempted to find and correct any bugs in this free software
// program, the author is not responsible for any damage or losses of any
// kind caused by the use or misuse of the program. You can redistribute
// the program and or modify it in any form without obligations, but the
// author would appreciated if these credits stays in.
//
// 
//

module decode_error
#( parameter ADRS_BITS = 32,
             DATA_BITS = 32,
             LEN_BITS  = 4,
             ID_BITS   = 6
 )             
(
   input              clk    ,
   input              reset_n,
   
   //
   // AXI port 
   //
   
   // write address 
   
   input                    awvalid , 
   output                   awready , 
   input    [ADRS_BITS-1:0] awaddr  , 
   input     [LEN_BITS-1:0] awlen   , 
   input             [ 2:0] awsize  , 
   input      [ID_BITS-1:0] awid    , 
   input             [ 1:0] awburst , 
                           
   // write data           
   input                    wvalid  , 
   output                   wready  , 
   input    [DATA_BITS-1:0] wdata   , 
   input  [DATA_BITS/8-1:0] wstrb   , 
   input                    wlast   , 
   input      [ID_BITS-1:0] wid     , 

   // write response    
   output reg               bvalid  , 
   input                    bready  , 
   output           [ 1 :0] bresp   , 
   output reg [ID_BITS-1:0] bid     , 


   //
   // read interface
   //
   
   // read address 
   input                    arvalid, 
   output                   arready, 
   input    [ADRS_BITS-1:0] araddr , 
   input      [ID_BITS-1:0] arid   , 
   input            [ 3: 0] arlen  , 
   input            [ 2: 0] arsize , 
   input            [ 1: 0] arburst, 

   // read data
   output reg               rvalid , 
   input                    rready , 
   output  [DATA_BITS-1:0]  rdata  , 
   output reg               rlast  , 
   output           [ 1 :0] rresp  , 
   output reg [ID_BITS-1:0] rid    ,
   
   // Give a pulse if an error address is seen
   output reg               fault                        
);

// shortcut for {LEN_BITS{1'b0}}
localparam [LEN_BITS-1:0] LEN0 = {LEN_BITS{1'b0}};
localparam [LEN_BITS-1:0] LEN1 = {{(LEN_BITS-1){1'b0}},1'b1};

reg [LEN_BITS-1:0] r_count;  // read retunr data burst count

   //
   // Write bus
   // 
   // Can accept one write address at a time
   // until the bresponse is gone (or going) as we need the ID
   // 
   assign awready = !bvalid | (bvalid & bready);
   
   // Straight forward sink all arriving data 
   assign wready  = 1'b1;
   assign bresp   = 2'b11; // Decode error
   
   
   //
   // Read bus 
   //
      
   // Can accept one read address at a time
   // and not until all previous data has been sent (or is leaving)
   assign arready = !rvalid | (rvalid & rready & rlast);
   
   assign rdata  = {DATA_BITS{1'b0}};
   assign rresp  = 2'b11; // Decode error

   //
   //
   //
   always @(posedge clk or negedge reset_n)
   begin
      if (!reset_n)
      begin
         bvalid  <= 1'b0;
         rvalid  <= 1'b0;
         fault   <= 1'b0;
      end
      else
      begin
         //
         // Write bus
         // use the address to give bresponse
         // Don't acare about data at all 
         //
         if (awvalid & awready)
         begin
            bid     <= awid;
            bvalid  <= 1'b1;
         end
         else
         begin
            if (bvalid & bready)
               bvalid  <= 1'b0;
         end
               
        
         //
         // Read bus
         //
         if (arvalid & arready)
         begin
            r_count <= arlen;
            rid     <= arid;
            rvalid  <= 1'b1;
            if (arlen!=LEN0)
               rlast <= 1'b0;
            else
               rlast <= 1'b1;
         end
         else
         begin
            // arready is never asserted in the middle of a burst
            // But it may be asserted in the last cycle!
            if (rvalid & rready)
            begin
               r_count <= r_count - 1;
               if (r_count==LEN1)
                  rlast <= 1'b1; 
               // Note the this rvalid assignment may be overriden
               // by the one above. This allows back-to-back AXI cycles
               if (rlast)
                  rvalid <= 1'b0;
            end
         end
         

         // use registered fault as the signal
         // may have to travel a long distance
         if ((awvalid & awready) | (arvalid & arready))
            fault <= 1'b1;
         else
            fault <= 1'b0;

      end
   end

endmodule
   